"""GPT4All Client."""
import logging
from dataclasses import dataclass, field
from typing import Optional

import llm

logger = logging.getLogger(__name__)

# GPT4ALL support is optional, so 
# if it's not installed, we'll just skip it
# Normally the __init__ gets this
# So this bit is here in case we try a direct import
try:
    from llm_gpt4all import Gpt4AllModel
except ModuleNotFoundError:
    logger.warning("llm_gpt4all module not found. GPT4All support will be disabled.")

@dataclass
class GPT4AllClient:
    """A client for interfacing with local GPT4All models.

    This interfaces with the llm-gpt4all module.
    """

    model: str = field(default_factory=lambda: "orca-mini-3b-gguf2-q4_0")
    cache_db_path: str = ""
    api_key: str = ""

    # TODO: Explicitly deprecate interactive mode for local models
    interactive: Optional[bool] = None

    local_model: Gpt4AllModel = None

    def __post_init__(self):
        logging.info(f"Preparing {self.model} for local use...")
        if self.model not in llm.get_model_aliases():
            raise ValueError(f"Model {self.model} not found in llm.get_model_aliases()")
        print(llm.get_model_aliases())
        self.local_model = llm.get_model(self.model)

    # Use this command in subsequent calls
    def chain_gpt4all_model(model, prompt_text):
        """Interact with a GPT4All model."""
        raw_output = model.prompt(prompt_text)

        return raw_output.text()

    # TODO: Dynamically update max_tokens
    # For now, this argument is ignored
    def complete(self, prompt, max_tokens=3000, show_prompt: bool = False, **kwargs) -> str:
        engine = self.model
        logger.info(f"Complete: engine={engine}, prompt[{len(prompt)}]={prompt[0:100]}...")
        if show_prompt:
            logger.info(f" SENDING PROMPT:\n{prompt}")

        response = None
        logger.debug("Prompting model for completion...")
        try:
            response = self.local_model.prompt(prompt)
        except Exception as e:
            logger.error(f"Encountered error: {e}")

        payload = response.text()

        return payload

    def db_connection(self):
        raise NotImplementedError("GPT4All local models do not currently support caching.")

    def cached_completions(self):
        raise NotImplementedError("GPT4All local models do not currently support caching.")

    def embeddings(self):
        raise NotImplementedError("GPT4All local models do not currently support embeddings.")

    def similarity(self):
        raise NotImplementedError("GPT4All local models do not currently support similarity.")

    def euclidian_distance(self):
        raise NotImplementedError(
            "GPT4All local models do not currently support distance measurement."
        )



from __future__ import annotations
from datetime import datetime, date
from enum import Enum

from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, ConfigDict,  Field, field_validator
import re
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        extra = 'forbid',
        arbitrary_types_allowed=True,
        use_enum_values = True)
    pass


class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    
    

class NamedEntity(ConfiguredBaseModel):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class DiagnosticProcedure(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Phenotype(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class ClinicalAttribute(NamedEntity):
    
    unit: Optional[str] = Field(None, description="""the unit used to measure the attribute""")
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Quality(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Unit(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    
    

class DiagnosticProceduretoPhenotypeAssociation(Triple):
    """
    A triple representing a relationship between a diagnostic procedure and an associated phenotype, e.g., \"blood pressure measurement\" is associated with \"high blood pressure\".
    """
    subject: Optional[str] = Field(None, description="""A diagnostic procedure yielding a result, which in turn may be interpreted as a phenotype. Procedures include \"heart rate measurement\", \"blood pressure measurement\", \"oxygen saturation measurement\", etc. In practice, procedures may be named based on what they measure, with the \"measurement\" part left implicit.""")
    predicate: Optional[str] = Field(None, description="""The relationship type, e.g. RELATED_TO""")
    object: Optional[List[str]] = Field(default_factory=list, description="""The observable physical or biochemical characteristics of a patient. Not equivalent to a disease state, but may contribute to a diagnosis.""")
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the procedure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the phenotype.""")
    
    

class DiagnosticProceduretoAttributeAssociation(Triple):
    """
    A triple representing a relationship between a diagnostic procedure and a measured attribute, e.g., \"blood pressure measurement\" is associated with \"blood pressure\" (or in OBA, something like OBA:VT0000183, \"blood pressure trait\").
    """
    subject: Optional[str] = Field(None, description="""A diagnostic procedure yielding a result, which in turn may be interpreted as a phenotype. Procedures include \"heart rate measurement\", \"blood pressure measurement\", \"oxygen saturation measurement\", etc. In practice, procedures may be named based on what they measure, with the \"measurement\" part left implicit.""")
    predicate: Optional[str] = Field(None, description="""The relationship type, e.g. RELATED_TO""")
    object: Optional[List[str]] = Field(default_factory=list, description="""Any measurable clinical attribute.""")
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the procedure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the phenotype.""")
    
    

class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    
    

class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    publication: Optional[Publication] = Field(None)
    entities: Optional[List[str]] = Field(default_factory=list)
    
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class ProcedureToPhenotypePredicate(RelationshipType):
    """
    A predicate for procedure to phenotype relationships, defining \"this procedure is intended to provide support for/against this phenotype\".
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class ProcedureToAttributePredicate(RelationshipType):
    """
    A predicate for procedure to attribute relationships, defining \"this procedure is a measurement of this attribute\".
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
DiagnosticProcedure.model_rebuild()
Phenotype.model_rebuild()
ClinicalAttribute.model_rebuild()
Quality.model_rebuild()
Unit.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
DiagnosticProceduretoPhenotypeAssociation.model_rebuild()
DiagnosticProceduretoAttributeAssociation.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
ProcedureToPhenotypePredicate.model_rebuild()
ProcedureToAttributePredicate.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()




"""Settings specific for using OpenAI models on the Azure platform."""

import logging
import os
from pathlib import Path
from typing import Any, Dict

import dpath
import toml

BASE_DIR = Path(__file__).parent.parent.absolute()

logger = logging.getLogger(__name__)

def read_toml(path: Path) -> Dict[str, Any]:
    with path.open() as cf:
        return toml.load(cf)


def merge_settings(overrides: Dict[str, Any], defaults: Dict[str, Any]) -> None:
    """Merge settings recursively to allow partial overrides without overwriting entire sections."""
    for key, value in overrides.items():
        if isinstance(value, dict):
            node = defaults.setdefault(key, {})
            merge_settings(value, node)
        else:
            defaults[key] = value


def parse_settings(
    base_dir: Path,
) -> Dict[str, Any]:
    """Parse configuration toml files from a directory with overrides.

    Args:
    base_dir: Base directory
    Returns: Dictionary of parsed configuration values.
    """
    try:
        settings_path = Path(os.getenv("SETTINGS_PATH", None))
        settings_name = os.getenv("SETTINGS_NAME", "local")
    except TypeError as e:
        logger.warning(f"Missing required settings for Azure OpenAI: {e}")
        settings_path = None
        settings_name = "local"

    etc_dir = base_dir / "etc"
    default_settings = etc_dir / f"{settings_name}.toml"
    settings_override = etc_dir / f"{settings_name}_custom.toml"

    if settings_path:
        settings = read_toml(settings_path)
    else:
        settings_path = default_settings

    settings = read_toml(settings_path) if settings_path.is_file() else {}

    return settings

settings_val = parse_settings(BASE_DIR)
try:
    AZURE_MODEL = dpath.get(settings_val, ["openai", "azure_deployment_name"])
    AZURE_API_VERSION = dpath.get(settings_val, ["openai", "azure_api_version"])
    AZURE_ENDPOINT = dpath.get(settings_val, ["openai", "azure_api_base"])
except KeyError as e:
    AZURE_MODEL = ""
    AZURE_API_VERSION = ""
    AZURE_ENDPOINT = ""
    logger.warning(f"Missing required settings for Azure OpenAI: {e}")


"""Loader functions for extraction templates."""

import importlib
import logging
from pathlib import Path

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import ClassDefinition

TEMPLATE_NAME = str

this_path = Path(__file__).parent
logger = logging.getLogger(__name__)


def get_template_details(template: TEMPLATE_NAME) -> ClassDefinition:
    """
    Get the LinkML class, module, and schemaview for a template.

    :param template: template name of the form module.ClassName
    :return: tuple of (LinkML class definition, module, specific
    Python class, schemaview)
    """
    logger.info(f"Loading schema for {template}")
    if "." in template:
        module_name, class_name = template.split(".", 1)
    else:
        module_name = template
        class_name = None
    path_to_template = get_template_path(module_name)
    sv = SchemaView(path_to_template)
    if class_name is None:
        roots = [c.name for c in sv.all_classes().values() if c.tree_root]
        if len(roots) != 1:
            raise ValueError(f"Template {template} does not have singular root: {roots}")
        class_name = roots[0]
    mod = importlib.import_module(f"ontogpt.templates.{module_name}")
    pyclass = mod.__dict__[class_name]

    logger.info(f"Getting class for template {template}")
    cls = None
    for c in sv.all_classes().values():
        if c.name == class_name:
            cls = c
            break
    if not cls:
        raise ValueError(f"Template {template} not found")
    return (cls, mod, pyclass, sv)


def get_template_path(module_name: str) -> Path:
    """Get the path to the YAML file for the template."""
    templates_path = this_path.parent / "templates"
    return templates_path / f"{module_name}.yaml"



"""OpenAI client."""

import ast
import logging
import sqlite3
import sys
from dataclasses import dataclass, field
from pathlib import Path
from time import sleep
from typing import Iterator, Optional, Tuple

import numpy as np
from openai import OpenAI

from oaklib.utilities.apikey_manager import get_apikey_value
from openai import AzureOpenAI

logger = logging.getLogger(__name__)

NUM_RETRIES = 3


@dataclass
class OpenAIClient:
    # max_tokens: int = field(default_factory=lambda: 3000)
    model: str = field(default_factory=lambda: "gpt-3.5-turbo")
    cache_db_path: str = ""
    api_key: str = ""
    interactive: Optional[bool] = None
    use_azure: Optional[bool] = None

    def __post_init__(self):
        if not self.api_key:
            self.api_key = get_apikey_value("openai")

        if self.use_azure:
            from ontogpt.utils.azure_settings import AZURE_MODEL, AZURE_API_VERSION, AZURE_ENDPOINT
            self.model = field(default_factory=lambda: AZURE_MODEL)
            # TODO: control client (Azure vs not) using a feature flag
            self.client = AzureOpenAI(
                api_version=AZURE_API_VERSION,
                azure_endpoint=AZURE_ENDPOINT,
                api_key=self.api_key,
                azure_deployment=AZURE_MODEL,
            )
        else:
            self.client = OpenAI(api_key=self.api_key)

    # TODO: Dynamically update max_tokens
    def complete(self, prompt, max_tokens=500, show_prompt: bool = False, **kwargs) -> str:
        # TODO: dynamically set model at call time
        engine = self.model

        logger.info(f"Complete: engine={engine}, prompt[{len(prompt)}]={prompt[0:100]}...")
        if show_prompt:
            logger.info(f" SENDING PROMPT:\n{prompt}")
        cur = self.db_connection()
        res = cur.execute("SELECT payload FROM cache WHERE prompt=? AND engine=?", (prompt, engine))
        payload = res.fetchone()
        if payload:
            prompt_peek = str(prompt)[0:80].replace("\n", "\\n")
            logger.info(f"Using cached payload for prompt: {prompt_peek}...")
            return payload[0]
        response = None
        i = 0
        while not response:
            i += 1
            logger.debug(f"Calling OpenAI API (attempt {i})...")
            try:
                if self.interactive:
                    response = self._interactive_completion(prompt, engine, max_tokens, **kwargs)
                elif self._must_use_chat_api() and self.use_azure:
                    response = self.client.chat.completions.create(
                        model=engine,
                        messages=[
                            {
                                "role": "user",
                                "content": prompt,
                            },
                        ],
                        max_tokens=max_tokens,
                        **kwargs,
                    )
                elif self._must_use_chat_api() and not self.use_azure:
                    response = self.client.chat.completions.create(model=engine,
                    messages=[
                        {
                            "role": "user",
                            "content": prompt,
                        },
                    ],
                    max_tokens=max_tokens,
                    **kwargs)                    
                else:
                    # TODO: remove chat api flag and eliminate references to legacy completions API
                    raise ValueError("Unsupported mode")
                break
            except Exception as e:
                logger.error(f"OpenAI API connection error: {e}")
                if i >= NUM_RETRIES:
                    raise e
                sleep_time = 4**i
                logger.info(f"Retrying {i} of {NUM_RETRIES} after {sleep_time} seconds...")
                sleep(sleep_time)

        #response = response.dict()
        #print(response)
        if self.interactive:
            payload = response
        elif self._must_use_chat_api():
            payload = response.choices[0].message.content
        else:
            payload = response.choices[0].text
        logger.info(f"Storing payload of len: {len(payload)}")
        cur.execute(
            "INSERT INTO cache (prompt, engine, payload) VALUES (?, ?, ?)",
            (prompt, engine, payload),
        )
        cur.connection.commit()
        return payload

    def db_connection(self):
        if not self.cache_db_path:
            self.cache_db_path = ".openai_cache.db"
        logger.info(f"Caching OpenAI responses to {Path(self.cache_db_path).absolute()}")
        create = not Path(self.cache_db_path).exists()
        con = sqlite3.connect(self.cache_db_path)
        cur = con.cursor()
        if create:
            cur.execute("CREATE TABLE cache (prompt, engine, payload)")
        return cur

    def _interactive_completion(
        self, prompt: str, engine: str, max_tokens: Optional[int], **kwargs
    ):
        print("Please use the ChatGPT interface to complete the following prompt:")
        print(f"IMPORTANT: make sure model == {engine}")
        print(f"Note: max_tokens == {max_tokens}")
        print("Prompt:")
        print(prompt)
        print("Paste in the response here, then ctrl-d to continue:")
        response = sys.stdin.read()
        print("OK? (y/n)")
        ok = input()
        if ok == "y":
            print("Thank you! This will now be cached.")
            print("Please be patient for the rest of the process to finish...")
            return response
        else:
            return self._interactive_completion(prompt, engine, max_tokens, **kwargs)

    def cached_completions(
        self, search_term: str = "", engine: str = ""
    ) -> Iterator[Tuple[str, str, str]]:
        if search_term:
            search_term = search_term.lower()
        cur = self.db_connection()
        res = cur.execute("SELECT engine, prompt, payload FROM cache")
        for row in res:
            if (
                search_term
                and search_term not in row[1].lower()
                and search_term not in row[2].lower()
            ):
                continue
            if engine and engine != row[0]:
                continue
            yield row

    def _must_use_chat_api(self) -> bool:
        """Return True if the model requires the chat API, False otherwise."""
        if self.model.startswith("text-davinci"):
            return False
        return True

    def embeddings(self, text: str, model: str = ""):
        text = str(text)

        if model == "":
            model = "text-embedding-ada-002"
        cur = self.db_connection()
        try:
            logger.info("creating embeddings cache")
            cur.execute("CREATE TABLE embeddings_cache (text, engine, vector_as_string)")
        except sqlite3.OperationalError:
            logger.info("Embeddings cache table already exists")
            pass
        res = cur.execute(
            "SELECT vector_as_string FROM embeddings_cache WHERE text=? AND engine=?", (text, model)
        )
        payload = res.fetchone()
        if payload:
            logger.info(f"Using cached embeddings for {model} {text[0:80]}...")
            return ast.literal_eval(payload[0])
        logger.info(f"querying OpenAI for {model} {text[0:80]}...")
        response = self.client.embeddings.create(model=model,
        input=text)
        v = response.data[0].embedding
        logger.info(f"Storing embeddings of len: {len(v)}")
        cur.execute(
            "INSERT INTO embeddings_cache (text, engine, vector_as_string) VALUES (?, ?, ?)",
            (text, model, str(v)),
        )
        cur.connection.commit()
        return v

    def similarity(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        logger.debug(f"similarity: {a1[0:10]}... x {a2[0:10]}... // ({len(a1)} x {len(a2)})")
        return np.dot(a1, a2) / (np.linalg.norm(a1) * np.linalg.norm(a2))

    def euclidian_distance(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        return np.linalg.norm(np.array(a1) - np.array(a2))



"""Clients for accessing different APIs."""
import logging
from ontogpt.clients.openai_client import OpenAIClient  # noqa:F401

# from ontogpt.clients.hfhub_client import HFHubClient  # noqa:F401
from ontogpt.clients.pubmed_client import PubmedClient  # noqa:F401
from ontogpt.clients.soup_client import SoupClient  # noqa:F401

# GPT4ALL support is optional, so 
# if it's not installed, we'll just skip it
try:
    from llm_gpt4all import Gpt4AllModel # noqa:F401
    from ontogpt.clients.gpt4all_client import GPT4AllClient  # noqa:F401
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.warning("llm_gpt4all module not found. GPT4All support will be disabled.")
